import sys

import pytest

from rgbkb.acer_ph16_71.device import AcerPredatorPH1671
from rgbkb.cli import CliParser


@pytest.mark.parametrize(
    "argv",
    [
        "static_color red",
        "static_color green",
        "static_color blue",
        "static_color magenta",
        "static_color yellow",
        "static_color cyan",
        "static_color white",
        "static_color on",
        "static_color black",
        "static_color off",
        "static_color fff",
        "static_color ffffff",
        "static_color f00",
        "static_color ff0000",
        "static_color white 100%",
        "static_color white 10%",
        "static_color white 5%",
        "static_color white 1",
        "static_color white 1.0",
        "static_color white 0.1",
        "static_color white .1",
        "static_color white 0.05",
        "static_color white .05",
    ],
)
def test_parser_with_one_command(argv):
    parser = CliParser(AcerPredatorPH1671)
    parsed = parser.parse(argv.split())
    print(parsed)
    assert parsed != ""


@pytest.mark.parametrize(
    "argv, expected",
    [
        ["flash_3x red", [["flash_3x", "red"]]],
        ["wave 33% fast right", [["wave", "33%","fast", "right"]]],
        ["wave .5 bright up", [["wave", ".5", "bright", "up"]]],
        ["static_color blue wasd_keys red", [["static_color", "blue"], ["wasd_keys", "red"]]],
        ["flash_3x magenta", [["flash_3x", "magenta"]]],
        ["static_color yellow", [["static_color", "yellow"]]],
        ["static_color cyan", [["static_color", "cyan"]]],
        ["static_color white", [["static_color", "white"]]],
        ["static_color on", [["static_color", "on"]]],
        ["static_color black", [["static_color", "black"]]],
        ["static_color off", [["static_color", "off"]]],
        ["static_color fff", [["static_color", "fff"]]],
        ["static_color ffffff", [["static_color", "ffffff"]]],
        ["static_color f00", [["static_color", "f00"]]],
        ["snow ff0000", [["snow", "ff0000"]]],
        ["snow crimson", [["snow", "crimson"]]],
        ["static_color white 100%", [["static_color", "white", "100%"]]],
        ["static_color white 10%", [["static_color", "white", "10%"]]],
        ["static_color white 5%", [["static_color", "white", "5%"]]],
        ["static_color white 1", [["static_color", "white", "1"]]],
        ["static_color white 1.0", [["static_color", "white", "1.0"]]],
        ["static_color white 0.1", [["static_color", "white", "0.1"]]],
        ["static_color white .1", [["static_color", "white", ".1"]]],
        ["static_color white 0.05", [["static_color", "white", "0.05"]]],
        ["static_color white .05", [["static_color", "white", ".05"]]],
   ],
)
def test_parser_with_more_commands(argv, expected):
    parser = CliParser(AcerPredatorPH1671)
    parsed = parser.parse(argv.split())
    print(parsed)
    assert parsed == expected


@pytest.mark.parametrize(
    "argv",
    [
        "the_thing red blue white",
        "white static .05",
    ],
)
def test_parser_should_fail_to_parse(argv):
    parser = CliParser(AcerPredatorPH1671)
    with pytest.raises(ValueError, match=r'Unknown keyword .+ on position \d+.'):
        parser.parse(argv.split())



import argparse
from textwrap import wrap
from usbx import Device, RequestType, Recipient, ControlTransfer, TransferDirection
from rgbkb.acer_ph16_71.commands import ARGUMENTS_TO_COMMANDS, EffectNames
from rgbkb.acer_ph16_71.keys import KeyNames, GroupNames
from rgbkb.kb import RgbKeyboard, KeyboardCommand

class OrderedArgs(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        if not 'ordered_args' in namespace:
            setattr(namespace, 'ordered_args', [])
        previous = namespace.ordered_args
        previous.append((self.dest, values))
        setattr(namespace, 'ordered_args', previous)

def add_parsers_to_cli_subparser(subparsers):
    # effects
    heartbeat_parser = subparsers.add_parser(
        name='heartbeat',
        help=f"help for heartbeat command."
    )
    heartbeat_parser.add_argument(
        'speed',
        nargs='?',
        type=str,
        metavar='NUMERIC VALUE'
    )
    heartbeat_parser.add_argument(
        'brightness',
        nargs='?',
        type=str,
        metavar='NUMERIC VALUE'
    )
    heartbeat_parser.add_argument(
        'color',
        nargs='?',
        type=str,
        metavar='COLOR',
        help="Accepts 'random', preset colors , named html colors, colors in the form of RGB ex.:1fe or RRGGBB ex.: 15f0e3"
    )
    snow_parser = subparsers.add_parser(
        name='snow',
        help=f"help for snow command."
    )
    snow_parser.add_argument(
        'speed',
        nargs='?',
        type=str,
        metavar='NUMERIC VALUE'
    )
    snow_parser.add_argument(
        'brightness',
        nargs='?',
        type=str,
        metavar='NUMERIC VALUE'
    )
    snow_parser.add_argument(
        'color',
        nargs='?',
        type=str,
        metavar='COLOR',
        help="Accepts 'random', preset colors , named html colors, colors in the form of RGB ex.:1fe or RRGGBB ex.: 15f0e3"
    )
    fireball_parser = subparsers.add_parser(
        name='fireball',
        help=f"help for fireball command."
    )
    fireball_parser.add_argument(
        'speed',
        nargs='?',
        type=str,
        metavar='NUMERIC VALUE'
    )
    fireball_parser.add_argument(
        'brightness',
        nargs='?',
        type=str,
        metavar='NUMERIC VALUE'
    )
    fireball_parser.add_argument(
        'color',
        nargs='?',
        type=str,
        metavar='COLOR',
        help="Accepts 'random', preset colors , named html colors, colors in the form of RGB ex.:1fe or RRGGBB ex.: 15f0e3"
    )
    stars_parser = subparsers.add_parser(
        name='stars',
        help=f"help for stars command."
    )
    stars_parser.add_argument(
        'speed',
        nargs='?',
        type=str,
        metavar='NUMERIC VALUE'
    )
    stars_parser.add_argument(
        'brightness',
        nargs='?',
        type=str,
        metavar='NUMERIC VALUE'
    )
    stars_parser.add_argument(
        'color',
        nargs='?',
        type=str,
        metavar='COLOR',
        help="Accepts 'random', preset colors , named html colors, colors in the form of RGB ex.:1fe or RRGGBB ex.: 15f0e3"
    )
    spot_parser = subparsers.add_parser(
        name='spot',
        help=f"help for spot command."
    )
    spot_parser.add_argument(
        'speed',
        nargs='?',
        type=str,
        metavar='NUMERIC VALUE'
    )
    spot_parser.add_argument(
        'brightness',
        nargs='?',
        type=str,
        metavar='NUMERIC VALUE'
    )
    spot_parser.add_argument(
        'color',
        nargs='?',
        type=str,
        metavar='COLOR',
        help="Accepts 'random', preset colors , named html colors, colors in the form of RGB ex.:1fe or RRGGBB ex.: 15f0e3"
    )
    lightning_parser = subparsers.add_parser(
        name='lightning',
        help=f"help for lightning command."
    )
    lightning_parser.add_argument(
        'speed',
        nargs='?',
        type=str,
        metavar='NUMERIC VALUE'
    )
    lightning_parser.add_argument(
        'brightness',
        nargs='?',
        type=str,
        metavar='NUMERIC VALUE'
    )
    lightning_parser.add_argument(
        'color',
        nargs='?',
        type=str,
        metavar='COLOR',
        help="Accepts 'random', preset colors , named html colors, colors in the form of RGB ex.:1fe or RRGGBB ex.: 15f0e3"
    )
    rain_parser = subparsers.add_parser(
        name='rain',
        help=f"help for rain command."
    )
    rain_parser.add_argument(
        'speed',
        nargs='?',
        type=str,
        metavar='NUMERIC VALUE'
    )
    rain_parser.add_argument(
        'brightness',
        nargs='?',
        type=str,
        metavar='NUMERIC VALUE'
    )
    rain_parser.add_argument(
        'color',
        nargs='?',
        type=str,
        metavar='COLOR',
        help="Accepts 'random', preset colors , named html colors, colors in the form of RGB ex.:1fe or RRGGBB ex.: 15f0e3"
    )
    neon_parser = subparsers.add_parser(
        name='neon',
        help=f"help for neon command."
    )
    neon_parser.add_argument(
        'speed',
        nargs='?',
        type=str,
        metavar='NUMERIC VALUE'
    )
    neon_parser.add_argument(
        'brightness',
        nargs='?',
        type=str,
        metavar='NUMERIC VALUE'
    )
    neon_parser.add_argument(
        'color',
        nargs='?',
        type=str,
        metavar='COLOR',
        help="Accepts 'random', preset colors , named html colors, colors in the form of RGB ex.:1fe or RRGGBB ex.: 15f0e3"
    )
    ripple_parser = subparsers.add_parser(
        name='ripple',
        help=f"help for ripple command."
    )
    ripple_parser.add_argument(
        'speed',
        nargs='?',
        type=str,
        metavar='NUMERIC VALUE'
    )
    ripple_parser.add_argument(
        'brightness',
        nargs='?',
        type=str,
        metavar='NUMERIC VALUE'
    )
    ripple_parser.add_argument(
        'color',
        nargs='?',
        type=str,
        metavar='COLOR',
        help="Accepts 'random', preset colors , named html colors, colors in the form of RGB ex.:1fe or RRGGBB ex.: 15f0e3"
    )
    snake_parser = subparsers.add_parser(
        name='snake',
        help=f"help for snake command."
    )
    snake_parser.add_argument(
        'speed',
        nargs='?',
        type=str,
        metavar='NUMERIC VALUE'
    )
    snake_parser.add_argument(
        'brightness',
        nargs='?',
        type=str,
        metavar='NUMERIC VALUE'
    )
    snake_parser.add_argument(
        'color',
        nargs='?',
        type=str,
        metavar='COLOR',
        help="Accepts 'random', preset colors , named html colors, colors in the form of RGB ex.:1fe or RRGGBB ex.: 15f0e3"
    )
    wave_parser = subparsers.add_parser(
        name='wave',
        help=f"help for wave command."
    )
    wave_parser.add_argument(
        'speed',
        nargs='?',
        type=str,
        metavar='NUMERIC VALUE'
    )
    wave_parser.add_argument(
        'brightness',
        nargs='?',
        type=str,
        metavar='NUMERIC VALUE'
    )
    wave_parser.add_argument(
        'color',
        nargs='?',
        type=str,
        metavar='COLOR',
        help="Accepts 'random', preset colors , named html colors, colors in the form of RGB ex.:1fe or RRGGBB ex.: 15f0e3"
    )
    breathing_parser = subparsers.add_parser(
        name='breathing',
        help=f"help for breathing command."
    )
    breathing_parser.add_argument(
        'speed',
        nargs='?',
        type=str,
        metavar='NUMERIC VALUE'
    )
    breathing_parser.add_argument(
        'brightness',
        nargs='?',
        type=str,
        metavar='NUMERIC VALUE'
    )
    breathing_parser.add_argument(
        'color',
        nargs='?',
        type=str,
        metavar='COLOR',
        help="Accepts 'random', preset colors , named html colors, colors in the form of RGB ex.:1fe or RRGGBB ex.: 15f0e3"
    )
    static_color_parser = subparsers.add_parser(
        name='static_color',
        help=f"help for static_color command."
    )
    static_color_parser.add_argument(
        'speed',
        nargs='?',
        type=str,
        metavar='NUMERIC VALUE'
    )
    static_color_parser.add_argument(
        'brightness',
        nargs='?',
        type=str,
        metavar='NUMERIC VALUE'
    )
    static_color_parser.add_argument(
        'color',
        nargs='?',
        type=str,
        metavar='COLOR',
        help="Accepts 'random', preset colors , named html colors, colors in the form of RGB ex.:1fe or RRGGBB ex.: 15f0e3"
    )
    flash_3x_parser = subparsers.add_parser(
        name='flash_3x',
        help=f"help for flash_3x command."
    )
    flash_3x_parser.add_argument(
        'speed',
        nargs='?',
        type=str,
        metavar='NUMERIC VALUE'
    )
    flash_3x_parser.add_argument(
        'brightness',
        nargs='?',
        type=str,
        metavar='NUMERIC VALUE'
    )
    flash_3x_parser.add_argument(
        'color',
        nargs='?',
        type=str,
        metavar='COLOR',
        help="Accepts 'random', preset colors , named html colors, colors in the form of RGB ex.:1fe or RRGGBB ex.: 15f0e3"
    )

    # per key
    parser = subparsers.add_parser(
        name='per_key',
        help=f"Set color per group or per key.\n Example: rgbkb per_key  --all black --f_keys red --letters crimson --number_row 0f0 --numpad_numbers sea_green"
    )
    key_groups = parser.add_argument_group('Key Groups')
    for keyword in tuple(group_name.value for group_name in GroupNames):
        key_groups.add_argument(
            f"--{keyword}",
            metavar="COLOR",
            action=OrderedArgs,
            help="Color of this entire group of keys, Accepts: RGB, RRGGBB and named html colors."

        )
    keys = parser.add_argument_group('Individual Keys')
    for keyword in tuple(key_name.value for key_name in KeyNames):
        keys.add_argument(
            f"--{keyword}",
            metavar="COLOR",
            action=OrderedArgs,
            help=f"Set color of {keyword}, Accepts: RGB, RRGGBB and named html colors."
        )


class AcerPredatorPH1671(RgbKeyboard):
    VID = 0x04F2

    PID = 0x0117
    MODEL = "PH16-71"
    NAME = "Acer Predator PH16-71 Keyboard"

    @staticmethod
    def update_subparser(subparsers):
        add_parsers_to_cli_subparser(subparsers)


    @staticmethod
    def parse_arguments(argv: list[str]) -> list[KeyboardCommand]:
        keywords = AcerPredatorPH1671.available_cli_keywords()
        i = 0
        split_arguments = []
        while i < len(argv):
            if argv[i] not in keywords:
                raise ValueError(f'Unknown keyword {argv[i]} on position {i}.')
            split_arguments.append([argv[i]])
            i += 1
            while i < len(argv) and argv[i] not in keywords:
                split_arguments[-1].append(argv[i])
                i += 1
        output = []
        for name, *arguments in split_arguments:
            try:
                command = ARGUMENTS_TO_COMMANDS[name].parse_arguments(arguments)
                output.append(command)
            except KeyError as e:
                raise ValueError(f'Unknown command {name} with arguments {arguments}.') from e
        return output

    @staticmethod
    def available_cli_keywords() -> tuple[str, ...]:
        return tuple(effect_name.value for effect_name in EffectNames) + tuple(
            key_name.value for key_name in KeyNames) + tuple(group_name.value for group_name in GroupNames)

    def __init__(self, device: Device):
        self.device = device

    def send_control_transfer(self, request: int, value: int, index: int, data: bytes, request_type: RequestType,
                              recipient: Recipient):
        transfer = ControlTransfer(request_type, recipient, request, value, index)
        self.device.control_transfer_out(transfer, data)

    def send_commands(self, *commands: bytes):
        self.device.detach_standard_drivers()
        self.device.open()
        self.device.claim_interface(3)
        self.device.clear_halt(4, TransferDirection.OUT)
        for command in commands:
            self.send_control_transfer(9, 0x0300, 3, command, RequestType.CLASS, Recipient.INTERFACE)
        self.device.release_interface(3)
        self.device.close()
        self.device.attach_standard_drivers()

    def color_at_index(self, color: str, index: int):
        self.device.detach_standard_drivers()
        self.device.open()
        self.device.claim_interface(3)
        self.device.clear_halt(4, TransferDirection.OUT)
        self.send_control_transfer(
            9,
            0x0300,
            3,
            bytes.fromhex('12 00 00 08  00 00 00 e5'),
            RequestType.CLASS,
            Recipient.INTERFACE
        )
        payload = '00000000' * 16 * 8
        payload = payload[:index * 8] + '00' + color + payload[index * 8:]
        for chunk in wrap(payload, 64 * 2):
            self.device.transfer_out(4, bytes.fromhex(chunk))
        self.send_control_transfer(
            9,
            0x0300,
            3,
            bytes.fromhex('08 02 33 05  32 08 01 82'),
            RequestType.CLASS,
            Recipient.INTERFACE
        )
        self.device.release_interface(3)
        self.device.close()
        self.device.attach_standard_drivers()



from abc import ABC, abstractmethod
from typing import Iterable


class KeyboardCommand(ABC):
    @staticmethod
    @abstractmethod
    def parse_arguments(argv: Iterable[str]) -> 'KeyboardCommand':
        ...

class RgbKeyboard(ABC):
    """Usb-Connected RGB Keyboard"""

    VID:int
    PID:int
    MODEL:str
    MODEL_NAME:str

    @staticmethod
    @abstractmethod
    def available_cli_keywords()->tuple[str]:
        ...

    @staticmethod
    @abstractmethod
    def update_subparser(subparsers):
        ...

    @staticmethod
    @abstractmethod
    def parse_arguments(argv : Iterable[str]) -> tuple[KeyboardCommand]:
        ...

    @abstractmethod
    def send_commands(self, *commands: bytes) -> None:
        ...




from enum import Enum


class KeyNames(Enum):
    A_KEY = 'a_key'
    ALT_L = 'alt_l'
    ALT_R = 'alt_r'
    APOSTROPHE = 'apostrophe'
    B_KEY = 'b_key'
    BACKSLASH = 'backslash'
    BACKSPACE = 'backspace'
    BACKTICK = 'backtick'
    C_KEY = 'c_key'
    CAPS_LOCK = 'capslock'
    COMMA = 'comma'
    CTRL_L = 'ctrl_l'
    CTRL_R = 'ctrl_r'
    D_KEY = 'd_key'
    DASH = 'dash'
    DELETE = 'delete'
    DOT = 'dot'
    DOWN_ARROW = 'down_arrow'
    E_KEY = 'e_key'
    ENTER = 'enter'
    EQUAL = 'equal'
    ESCAPE = 'escape'
    F1 = 'f1'
    F2 = 'f2'
    F3 = 'f3'
    F4 = 'f4'
    F5 = 'f5'
    F6 = 'f6'
    F7 = 'f7'
    F8 = 'f8'
    F9 = 'f9'
    F10 = 'f10'
    F11 = 'f11'
    F12 = 'f12'
    F_KEY = 'f_key'
    FAST_FORWARD = 'fast_forward'
    FN = 'fn'
    G_KEY = 'g_key'
    H_KEY = 'h_key'
    I_KEY = 'i_key'
    INSERT = 'insert'
    J_KEY = 'j_key'
    K_KEY = 'k_key'
    KP_0 = 'kp_0'
    KP_1 = 'kp_1'
    KP_2 = 'kp_2'
    KP_3 = 'kp_3'
    KP_4 = 'kp_4'
    KP_5 = 'kp_5'
    KP_6 = 'kp_6'
    KP_7 = 'kp_7'
    KP_8 = 'kp_8'
    KP_9 = 'kp_9'
    KP_ASTERISK = 'kp_asterisk'
    KP_DOT = 'kp_dot'
    KP_ENTER = 'kp_enter'
    KP_FORWARD_SLASH = 'kp_forward_slash'
    KP_MINUS = 'kp_minus'
    KP_PLUS = 'kp_plus'
    L_KEY = 'l_key'
    LEFT_ARROW = 'left_arrow'
    LEFT_BRACKET = 'left_bracket'
    M_KEY = 'm_key'
    MENU = 'menu'
    N_KEY = 'n_key'
    NUM_LOCK = 'num_lock'
    NUMBER_ROW_0 = 'number_row_0'
    NUMBER_ROW_1 = 'number_row_1'
    NUMBER_ROW_2 = 'number_row_2'
    NUMBER_ROW_3 = 'number_row_3'
    NUMBER_ROW_4 = 'number_row_4'
    NUMBER_ROW_5 = 'number_row_5'
    NUMBER_ROW_6 = 'number_row_6'
    NUMBER_ROW_7 = 'number_row_7'
    NUMBER_ROW_8 = 'number_row_8'
    NUMBER_ROW_9 = 'number_row_9'
    O_KEY = 'o_key'
    P_KEY = 'p_key'
    PLAY_PAUSE = 'play_pause'
    POWER = 'power'
    PREDATOR = 'predator'
    PRINT_SCREEN = 'print_screen'
    Q_KEY = 'q_key'
    R_KEY = 'r_key'
    REWIND = 'rewind'
    RIGHT_ARROW = 'right_arrow'
    RIGHT_BRACKET = 'right_bracket'
    S_KEY = 's_key'
    SEMICOLON = 'semicolon'
    SHIFT_L = 'shift_l'
    SHIFT_R = 'shift_r'
    SLASH = 'slash'
    SPACE = 'space'
    T_KEY = 't_key'
    TAB = 'tab'
    U_KEY = 'u_key'
    UP_ARROW = 'up_arrow'
    V_KEY = 'v_key'
    W_KEY = 'w_key'
    WIN = 'win'
    X_KEY = 'x_key'
    Y_KEY = 'y_key'
    Z_KEY = 'z_key'


class GroupNames(Enum):
    ALL = "all"
    LETTERS = "letters"
    F_KEYS = "f_keys"
    NUMBER_ROW = "number_row"
    WASD_KEYS = "wasd_keys"
    ARROW_KEYS = "arrow_keys"
    NUMPAD_NUMBERS = "numpad_numbers"
    MEDIA_KEYS = "media_keys"
    MODIFIER_KEYS = "modifier_keys"


KEY_GROUPS: dict[GroupNames, tuple[KeyNames, ...]] = {
    GroupNames.ALL: tuple(KeyNames),
    GroupNames.LETTERS: (
        KeyNames.A_KEY,
        KeyNames.B_KEY,
        KeyNames.C_KEY,
        KeyNames.D_KEY,
        KeyNames.E_KEY,
        KeyNames.F_KEY,
        KeyNames.G_KEY,
        KeyNames.H_KEY,
        KeyNames.I_KEY,
        KeyNames.J_KEY,
        KeyNames.K_KEY,
        KeyNames.L_KEY,
        KeyNames.M_KEY,
        KeyNames.N_KEY,
        KeyNames.O_KEY,
        KeyNames.P_KEY,
        KeyNames.Q_KEY,
        KeyNames.R_KEY,
        KeyNames.S_KEY,
        KeyNames.T_KEY,
        KeyNames.U_KEY,
        KeyNames.V_KEY,
        KeyNames.W_KEY,
        KeyNames.X_KEY,
        KeyNames.Y_KEY,
        KeyNames.Z_KEY
    ),
    GroupNames.F_KEYS: (
        KeyNames.F1,
        KeyNames.F2,
        KeyNames.F3,
        KeyNames.F4,
        KeyNames.F5,
        KeyNames.F6,
        KeyNames.F7,
        KeyNames.F8,
        KeyNames.F9,
        KeyNames.F10,
        KeyNames.F11,
        KeyNames.F12
    ),
    GroupNames.NUMBER_ROW: (
        KeyNames.NUMBER_ROW_1,
        KeyNames.NUMBER_ROW_2,
        KeyNames.NUMBER_ROW_3,
        KeyNames.NUMBER_ROW_4,
        KeyNames.NUMBER_ROW_5,
        KeyNames.NUMBER_ROW_6,
        KeyNames.NUMBER_ROW_7,
        KeyNames.NUMBER_ROW_8,
        KeyNames.NUMBER_ROW_9,
        KeyNames.NUMBER_ROW_0
    )
    ,
    GroupNames.WASD_KEYS: (
        KeyNames.W_KEY,
        KeyNames.A_KEY,
        KeyNames.S_KEY,
        KeyNames.D_KEY
    ),
    GroupNames.ARROW_KEYS: (
        KeyNames.LEFT_ARROW,
        KeyNames.UP_ARROW,
        KeyNames.RIGHT_ARROW,
        KeyNames.DOWN_ARROW
    ),
    GroupNames.NUMPAD_NUMBERS: (
        KeyNames.KP_0,
        KeyNames.KP_1,
        KeyNames.KP_2,
        KeyNames.KP_3,
        KeyNames.KP_4,
        KeyNames.KP_5,
        KeyNames.KP_6,
        KeyNames.KP_7,
        KeyNames.KP_8,
        KeyNames.KP_9
    ),
    GroupNames.MEDIA_KEYS: (
        KeyNames.PLAY_PAUSE,
        KeyNames.REWIND,
        KeyNames.FAST_FORWARD
    ),
    GroupNames.MODIFIER_KEYS: (
        KeyNames.CTRL_L,
        KeyNames.ALT_L,
        KeyNames.SHIFT_L,
        KeyNames.CTRL_R,
        KeyNames.ALT_R,
        KeyNames.SHIFT_R
    )
}

# indexes with seemingly no keys:
# 7,8,9,14,30,36,42,48,54,66,79,
# 87,88,91.92.93,94,97,98,113,119,
# 121,125,126,127
KEYS: tuple[tuple[KeyNames, int, int, int], ...] = (
    # KeyName,Row, Col, PayloadIndex
    (KeyNames.CTRL_L, 0, 0, 0),
    (KeyNames.FN, 0, 1, 6),
    (KeyNames.WIN, 0, 2, 12),
    (KeyNames.ALT_L, 0, 3, 18),
    (KeyNames.SPACE, 0, 4, 24),
    (KeyNames.ALT_R, 0, 5, 60),
    (KeyNames.MENU, 0, 6, 72),
    (KeyNames.CTRL_R, 0, 7, 78),
    (KeyNames.LEFT_ARROW, 0, 8, 84),
    (KeyNames.DOWN_ARROW, 0, 9, 96),
    (KeyNames.RIGHT_ARROW, 0, 10, 102),
    (KeyNames.KP_0, 0, 11, 108),
    (KeyNames.KP_DOT, 0, 12, 114),
    (KeyNames.KP_ENTER, 0, 13, 120),
    (KeyNames.SHIFT_L, 1, 0, 1),
    (KeyNames.Z_KEY, 1, 1, 19),
    (KeyNames.X_KEY, 1, 2, 25),
    (KeyNames.C_KEY, 1, 3, 31),
    (KeyNames.V_KEY, 1, 4, 37),
    (KeyNames.B_KEY, 1, 5, 43),
    (KeyNames.N_KEY, 1, 6, 49),
    (KeyNames.M_KEY, 1, 7, 55),
    (KeyNames.COMMA, 1, 8, 61),
    (KeyNames.DOT, 1, 9, 67),
    (KeyNames.SLASH, 1, 10, 73),
    (KeyNames.SHIFT_R, 1, 11, 85),
    (KeyNames.UP_ARROW, 1, 12, 90),
    (KeyNames.KP_1, 1, 13, 100),
    (KeyNames.KP_2, 1, 14, 109),
    (KeyNames.KP_3, 1, 15, 115),
    (KeyNames.CAPS_LOCK, 2, 0, 2),
    (KeyNames.A_KEY, 2, 1, 20),
    (KeyNames.S_KEY, 2, 2, 26),
    (KeyNames.D_KEY, 2, 3, 32),
    (KeyNames.F_KEY, 2, 4, 38),
    (KeyNames.G_KEY, 2, 5, 44),
    (KeyNames.H_KEY, 2, 6, 50),
    (KeyNames.J_KEY, 2, 7, 56),
    (KeyNames.K_KEY, 2, 8, 62),
    (KeyNames.L_KEY, 2, 9, 68),
    (KeyNames.SEMICOLON, 2, 10, 74),
    (KeyNames.APOSTROPHE, 2, 11, 80),
    (KeyNames.ENTER, 2, 12, 86),
    (KeyNames.KP_4, 2, 13, 101),
    (KeyNames.KP_5, 2, 14, 110),
    (KeyNames.KP_6, 2, 15, 116),
    (KeyNames.KP_PLUS, 2, 16, 122),
    (KeyNames.TAB, 3, 0, 3),
    (KeyNames.Q_KEY, 3, 1, 15),
    (KeyNames.W_KEY, 3, 2, 21),
    (KeyNames.E_KEY, 3, 3, 27),
    (KeyNames.R_KEY, 3, 4, 33),
    (KeyNames.T_KEY, 3, 5, 39),
    (KeyNames.Y_KEY, 3, 6, 45),
    (KeyNames.U_KEY, 3, 7, 51),
    (KeyNames.I_KEY, 3, 8, 57),
    (KeyNames.O_KEY, 3, 9, 63),
    (KeyNames.P_KEY, 3, 10, 69),
    (KeyNames.LEFT_BRACKET, 3, 11, 75),
    (KeyNames.RIGHT_BRACKET, 3, 12, 81),
    (KeyNames.BACKSLASH, 3, 13, 13),
    (KeyNames.KP_7, 3, 14, 99),
    (KeyNames.KP_8, 3, 15, 111),
    (KeyNames.KP_9, 3, 16, 117),
    (KeyNames.KP_MINUS, 3, 17, 123),
    (KeyNames.BACKTICK, 4, 0, 4),
    (KeyNames.NUMBER_ROW_1, 4, 1, 10),
    (KeyNames.NUMBER_ROW_2, 4, 2, 16),
    (KeyNames.NUMBER_ROW_3, 4, 3, 22),
    (KeyNames.NUMBER_ROW_4, 4, 4, 28),
    (KeyNames.NUMBER_ROW_5, 4, 5, 34),
    (KeyNames.NUMBER_ROW_6, 4, 6, 40),
    (KeyNames.NUMBER_ROW_7, 4, 7, 46),
    (KeyNames.NUMBER_ROW_8, 4, 8, 52),
    (KeyNames.NUMBER_ROW_9, 4, 9, 58),
    (KeyNames.NUMBER_ROW_0, 4, 10, 64),
    (KeyNames.DASH, 4, 11, 70),
    (KeyNames.EQUAL, 4, 12, 76),
    (KeyNames.BACKSPACE, 4, 13, 82),
    (KeyNames.PREDATOR, 4, 14, 106),
    (KeyNames.NUM_LOCK, 4, 15, 112),
    (KeyNames.KP_FORWARD_SLASH, 4, 16, 118),
    (KeyNames.KP_ASTERISK, 4, 17, 124),
    (KeyNames.ESCAPE, 5, 0, 5),
    (KeyNames.F1, 5, 1, 11),
    (KeyNames.F2, 5, 2, 17),
    (KeyNames.F3, 5, 3, 23),
    (KeyNames.F4, 5, 4, 29),
    (KeyNames.F5, 5, 5, 35),
    (KeyNames.F6, 5, 6, 41),
    (KeyNames.F7, 5, 7, 47),
    (KeyNames.F8, 5, 8, 53),
    (KeyNames.F9, 5, 9, 59),
    (KeyNames.F10, 5, 10, 65),
    (KeyNames.F11, 5, 11, 71),
    (KeyNames.F12, 5, 12, 77),
    (KeyNames.PRINT_SCREEN, 5, 13, 83),
    (KeyNames.INSERT, 5, 14, 89),
    (KeyNames.DELETE, 5, 15, 95),
    (KeyNames.REWIND, 5, 16, 104),
    (KeyNames.PLAY_PAUSE, 5, 17, 105),
    (KeyNames.FAST_FORWARD, 5, 18, 103),
    (KeyNames.POWER, 5, 19, 107),
)



import argparse
import inspect
from collections.abc import Iterable
from typing import Type

from rgbkb.acer_ph16_71.commands import color, static_color
from rgbkb.utils import find_supported_devices
from rgbkb.kb import RgbKeyboard


def print_help():
    print(inspect.cleandoc("""
    usage: rgbkb [-h,--help] [-g GROUP GROUP] [-k KEY KEY] [color] effect [effect ...]
    RGB Keyboard Control (Written for a Acer Predator PH16-71, but might work for others too.
    Lives on GitHub @ https://github.com/fuho/rgbkb

    positional arguments:
      color                 Whole keyboard, one color
      effect                Animated effect, can have up to 4 sub-arguments

    options:
      -h, --help            show this help message and exit
      -g GROUP GROUP, --group GROUP GROUP
                            Group of keys, `-g function red` or `-g keypad 00ccff`
      -k KEY KEY, --key KEY KEY
                            One single key; `-k f_key white`

    I'd REALLY appreciate any feedback, especially devices you tried to use and if it worked or not.
    """))


class CliParser:
    def __init__(self, kb: Type[RgbKeyboard]):
        self.keywords = kb.available_cli_keywords()

    def parse(self, argv) -> Iterable[Iterable[str]]:
        i = 0
        commands = []
        while i < len(argv):
            if argv[i] not in self.keywords:
                raise ValueError(f'Unknown keyword {argv[i]} on position {i}.')
            commands.append([argv[i]])
            i += 1
            while i < len(argv) and argv[i] not in self.keywords:
                commands[-1].append(argv[i])
                i += 1
        return commands


def main():
    available_keyboards = find_supported_devices()
    argparser = argparse.ArgumentParser(
        prog='rgbkb',
        description='Controls keyboard backlight, specifically Acer Predator Helios 16',
        epilog="And that's the way the cookie crumbles."
    )
    group = argparser.add_mutually_exclusive_group()
    group.add_argument(
        '-l',
        '--list',
        nargs='?',
        type=bool,
        const=True,
        default=False,
        help='List all found compatible keyboards.'
    )
    group.add_argument(
        '-s',
        '--select',
        # metavar='DEVICE',
        nargs='?',
        type=int,
        const=0,
        default=0,
        choices=[i for i, _ in enumerate(available_keyboards)],
        help=f'Choose an ID from available supported keyboards: {','.join(f"{i}: {kb.MODEL}" for i, kb in enumerate(available_keyboards))}'
    )

    for kb in available_keyboards:
        kb.update_subparser(argparser.add_subparsers(dest="command"))
    ns = argparser.parse_args()
    print(ns)
    if ns.list:
        index = 0
        for kb in available_keyboards:
            print(f'{index} :  {kb.NAME} Model: {kb.MODEL} Vendor ID: {kb.VID:04x} Product ID: {kb.PID:04x}')
        return
    else:
        kb = available_keyboards[ns.select]

    # if len(sys.argv) <= 1:
    #     print_help()
    #     sys.exit(1)
    # elif sys.argv[1] == '-h' or sys.argv[1] == '--help':
    #     print_help()
    #     sys.exit(0)
    # else:
    #     parser = CliParser(AcerPredatorPH1671)
    #     parsed = parser.parse(sys.argv[1:])
    #     pprint(parsed)
    #     sys.exit(0)

    kb.send_commands(color(255, 255, 255), static_color(1, 32, 1, 0))
    # kb.send_commands(color(200, 100, 50), flash3x(0, 16, 1, 0))
    # kb.send_commands(flash3x(0, 16, 2, 1))
    # kb.send_commands('b1 00 00 00  00 00 00 4e','14 00 00 ff ff ff 00 ee', '08 02 29 ff 30 01 01 9b')
    # kb.send_commands('08 02 08 0b  19 01 00 00')



import sys

import pytest

from rgbkb.cli import main


def test_cli_no_arguments(monkeypatch, capsys):
    monkeypatch.setattr(sys, 'argv', ['rgbkb'])
    with pytest.raises(SystemExit) as pytest_e:
        main()
    captured = capsys.readouterr()
    assert captured.err == ''
    assert len(captured.err) == 0
    assert len(captured.out) > 0
    assert pytest_e.type == SystemExit
    assert pytest_e.value.code == 1
    assert 'usage: rgbkb' in captured.out


def test_cli_help(monkeypatch, capsys):
    monkeypatch.setattr(sys, 'argv', ['rgbkb', '-h'])
    with pytest.raises(SystemExit) as pytest_e:
        main()
    captured = capsys.readouterr()
    assert len(captured.err) == 0
    assert len(captured.out) > 0
    assert pytest_e.type == SystemExit
    assert pytest_e.value.code == 0
    assert 'usage: rgbk' in captured.out


@pytest.mark.parametrize(
    "argv",
    [
        "static_color red",
        "static_color green",
        "static_color blue",
        "static_color magenta",
        "static_color yellow",
        "static_color cyan",
        "static_color white",
        "static_color on",
        "static_color black",
        "static_color off",
        "static_color fff",
        "static_color ffffff",
        "static_color f00",
        "static_color ff0000",
        "static_color white 100%",
        "static_color white 10%",
        "static_color white 5%",
        "static_color white 1",
        "static_color white 1.0",
        "static_color white 0.1",
        "static_color white .1",
        "static_color white 0.05",
        "static_color white .05",
    ],
)
def test_cli_arguments_static_color(monkeypatch, capsys, argv):
    monkeypatch.setattr(sys, 'argv', ("rgbkb " + argv).split())
    with pytest.raises(SystemExit) as pytest_e:
        main()
    assert pytest_e.type == SystemExit
    assert pytest_e.value.code == 0



import binascii
from symtable import Class
from typing import Type
from usbx import Device, usb
from rgbkb.kb import RgbKeyboard
from rgbkb.acer_ph16_71.device import AcerPredatorPH1671

SUPPORTED_DEVICES: tuple[Type[RgbKeyboard]] = (
    AcerPredatorPH1671,
)


def chunkify(s, separator=" ", chunk_size=4):
    # Split the string into chunks of size `chunk_size`
    chunks = [s[i:i + chunk_size] for i in range(0, len(s), chunk_size)]
    # Join the chunks with separator and return
    formatted_string = separator.join(chunks)
    return formatted_string


def print_device_info(device: Device):
    print(
        f'VID: 0x{device.vid:04x}, PID: 0x{device.pid:04x}, Serial: {device.serial}, Protocol: {device.protocol_code}, Class: {device.class_code} {device.manufacturer}')
    print(f'Manufacturer: {device.manufacturer} Product: {device.product} Version: {device.device_version}')
    print(f'Configuration value: {device.configuration_value} Configuration descriptor:')
    print(chunkify(chunkify(binascii.hexlify(device.configuration_descriptor).decode('utf-8')), '\n', 80))
    print(f'{len(device.configuration.interfaces)} interfaces:')
    for interface in device.configuration.interfaces:
        print(f'\tInterface {interface.number} (currently selected alternate {interface.current_alternate.number})')
        for alternate in interface.alternates:
            print(
                f"\t\tAlternate {alternate.number} - Class:{alternate.class_code} SubClass:{alternate.subclass_code}"
                f" Protocol: {alternate.protocol_code} number of endpoints: {len(alternate.endpoints)}"
            )
            for endpoint in alternate.endpoints:
                print(
                    f"\t\t\tEndpoint {endpoint.number} Direction: {endpoint.direction.name}"
                    f" Transfer type: {endpoint.transfer_type.name} Max-Packet size: {endpoint.max_packet_size}b"
                )

def is_supported_device(device: Device) -> bool:
    if device is None or not isinstance(device, Device):
        return False
    known_devices = (f"{kb.VID}:{kb.PID}" for kb in SUPPORTED_DEVICES)
    return f"{device.vid}:{device.pid}" in known_devices


def find_supported_devices() -> tuple[RgbKeyboard,...]:
    known_devices = {f"{d.VID}:{d.PID}": d for d in SUPPORTED_DEVICES}
    available_known_devices = tuple(known_devices[f"{d.vid}:{d.pid}"].__call__(d) for d in usb.find_devices(is_supported_device))
    return available_known_devices


from enum import Enum

from rgbkb.kb import KeyboardCommand


class EffectNames(Enum):
    HEARTBEAT = 'heartbeat'
    SNOW = 'snow'
    FIREBALL = 'fireball'
    STARS = 'stars'
    SPOT = 'spot'
    LIGHTNING = 'lightning'
    RAIN = 'rain'
    NEON = 'neon'
    RIPPLE = 'ripple'
    SNAKE = 'snake'
    WAVE = 'wave'
    BREATHING = 'breathing'
    STATIC_COLOR = 'static_color'
    FLASH_3X = 'flash_3x'


effect_command = lambda identifier, *args: bytes.fromhex(identifier) + bytes((*args, 0))

color = lambda *args: effect_command('14 00 00', *args, 0)
heartbeat = lambda *args: effect_command('08 02 29', *args)
snow = lambda *args: effect_command('08 02 28', *args)
fireball = lambda *args: effect_command('08 02 27', *args)
stars = lambda *args: effect_command('08 02 26', *args)
spot = lambda *args: effect_command('08 02 25', *args)
lightning = lambda *args: effect_command('08 02 12', *args)
rain = lambda *args: effect_command('08 02 0a', *args)
neon = lambda *args: effect_command('08 02 08', *args)
ripple = lambda *args: effect_command('08 02 06', *args)
snake = lambda *args: effect_command('08 02 05', *args)
wave = lambda *args: effect_command('08 02 03', *args)
breathing = lambda *args: effect_command('08 02 02', *args)
static_color = lambda *args: effect_command('08 02 01', *args)
flash3x = lambda *args: effect_command('08 03 22', *args)

# @formatter:off
ALL_KEYS_WHITE = (
    ('send_control_request', 9, 0x0300, 3, bytes.fromhex('12 00 00 08  00 00 00 e5')),
    ('send_interrupt_transfer', 3, 4, bytes.fromhex('00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff')),
    ('send_interrupt_transfer', 3, 4, bytes.fromhex('00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff')),
    ('send_interrupt_transfer', 3, 4, bytes.fromhex('00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff')),
    ('send_interrupt_transfer', 3, 4, bytes.fromhex('00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff')),
    ('send_interrupt_transfer', 3, 4, bytes.fromhex('00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff')),
    ('send_interrupt_transfer', 3, 4, bytes.fromhex('00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff')),
    ('send_interrupt_transfer', 3, 4, bytes.fromhex('00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff')),
    ('send_interrupt_transfer', 3, 4, bytes.fromhex('00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff00ffffff0000000000000000')),
    ('send_control_request', 9, 0x0300, 3, bytes.fromhex('08 02 33 05  32 08 01 82')),
)
# @formatter:on

# @formatter:off
FLASH_KEYS = (
    # Captured
    ('14 00 01 2e  09 c7 00 ec', '08 03 22 05  32 02 01 98'),  # 3 flash blue
    ('14 00 01 c7  00 ff 00 24', '08 03 22 05  32 02 01 98'),  # 3 flash purple
    # Synthesised
    # 14 00 01 RR  GG BB 00 24
    ('14 00 01 ff  00 00 00 00', '08 03 22 05  32 02 01 98'),  # 3 flash red
    ('14 00 01 ff  ff 00 00 00', '08 03 22 05  32 02 01 98'),  # 3 flash yellow
    ('14 00 01 00  ff 00 00 00', '08 03 22 05  32 02 01 98'),  # 3 flash green
    ('14 00 01 00  ff ff 00 00', '08 03 22 05  32 02 01 98'),  # 3 flash cyan
    ('14 00 01 00  00 ff 00 00', '08 03 22 05  32 02 01 98'),  # 3 flash blue
    ('14 00 01 ff  00 ff 00 00', '08 03 22 05  32 02 01 98'),  # 3 flash purple
    ('14 00 01 ff  ff ff 00 00', '08 03 22 05  32 02 01 98'),  # 3 flash white
    (                            '08 03 22 01  00 00 01 00',),  # 3 flash cyan
    (                            '08 03 22 00  00 00 00 ff',),  # 3 flash cyan
    (                            '08 03 22 00  00 00 00 00',),  # 3 flash cyan
    (                            '08 03 22 ff  ff 00 ff ff',),  # 3 flash cyan
    (                            '08 03 22 00  00 02 00 00',),  # 3 flash red
    (                            '08 03 22 ff  ff 02 ff ff',),  # 3 flash red
    (                            '08 03 22 00  00 03 00 00',),  # 3 flash yellow
    (                            '08 03 22 ff  ff 03 ff ff',),  # 3 flash yellow
    (                            '08 03 22 ff  ff 04 ff ff',),  # 3 flash last set color
    (                            '08 03 22 00  00 04 00 00',),  # 3 flash last set color
    (                            '08 03 22 00  00 05 00 00',),  # 3 flash last set color
    (                            '08 03 22 00  00 06 00 00',),  # 3 flash last set color
    (                            '08 03 22 00  00 07 00 00',),  # 3 flash last set color
    (                            '08 03 22 ff  ff ff ff ff',),  # 3 flash last set color
    (                            '08 03 22 00  00 01 00 ff',),  # 3 flash last set color
    (                            '08 03 22 00  00 04 00 ff',),  # 3 flash last static color?
    (                            '08 03 22 00  00 05 00 ff',),  # 3 flash last static color?
    # Working on this one
    (                            '08 03 22 00  00 18 00 00',),  # 3 flash last set color
)

BRIGHTNESS_MAX = 0x32
SPEED_MAX = 0x1b

class Color:
    def __init__(self, red: int, green: int, blue: int):
        self.red = red % 256
        self.green = green % 256
        self.blue = blue % 256

class ColorPreset(Enum):
    STATIC = 1
    GREEN = 2
    YELLOW = 3
    BLUE = 3
    CYAN = 4
    MAGENTA = 5
    WHITE = 6
    RANDOM = 7

class HtmlColors(Enum):
    ALICE_BLUE = "f0f8ff"  # White
    ANTIQUE_WHITE = "faebd7"  # White
    AQUA = "00ffff"  # Blue
    AQUA_MARINE = "7fffd4"  # Blue
    AZURE = "f0ffff"  # White
    BEIGE = "f5f5dc"  # White
    BISQUE = "ffe4c4"  # Brown
    BLACK = "000000"  # Gray
    BLANCHED_ALMOND = "ffebcd"  # Brown
    BLUE = "0000ff"  # Blue (one of preset colors)
    BLUE_VIOLET = "8a2be2"  # Purple
    BROWN = "a52a2a"  # Brown
    BURLY_WOOD = "deb887"  # Brown
    CADET_BLUE = "5f9ea0"  # Blue
    CHARTREUSE = "7fff00"  # Green
    CHOCOLATE = "d2691e"  # Brown
    CORAL = "ff7f50"  # Orange
    CORN_FLOWER_BLUE = "6495ed"  # Blue
    CORNSILK = "fff8dc"  # Brown
    CRIMSON = "dc143c"  # Red
    CYAN = "00ffff"  # Blue
    DARK_BLUE = "00008b"  # Blue
    DARK_CYAN = "008b8b"  # Green
    DARK_GOLDEN_ROD = "b8860b"  # Brown
    DARK_GRAY = "a9a9a9"  # Gray
    DARK_GREEN = "006400"  # Green
    DARK_KHAKI = "bdb76b"  # Yellow
    DARK_MAGENTA = "8b008b"  # Purple
    DARK_OLIVE_GREEN = "556b2f"  # Green
    DARK_ORANGE = "ff8c00"  # Orange
    DARK_ORCHID = "9932cc"  # Purple
    DARK_RED = "8b0000"  # Red
    DARK_SALMON = "e9967a"  # Red
    DARK_SEA_GREEN = "8fbc8f"  # Green
    DARK_SLATE_BLUE = "483d8b"  # Purple
    DARK_SLATE_GRAY = "2f4f4f"  # Gray
    DARK_TURQUOISE = "00ced1"  # Blue
    DARK_VIOLET = "9400d3"  # Purple
    DEEP_PINK = "ff1493"  # Pink
    DEEP_SKY_BLUE = "00bfff"  # Blue
    DIM_GRAY = "696969"  # Gray
    DODGER_BLUE = "1e90ff"  # Blue
    FIRE_BRICK = "b22222"  # Red
    FLORAL_WHITE = "fffaf0"  # White
    FOREST_GREEN = "228b22"  # Green
    FUCHSIA = "ff00ff"  # Purple
    GAINSBORO = "dcdcdc"  # Gray
    GHOST_WHITE = "f8f8ff"  # White
    GOLD = "ffd700"  # Yellow
    GOLDEN_ROD = "daa520"  # Brown
    GRAY = "808080"  # Gray
    GREEN = "008000"  # Green
    GREEN_YELLOW = "adff2f"  # Green
    HONEY_DEW = "f0fff0"  # White
    HOT_PINK = "ff69b4"  # Pink
    INDIAN_RED = "cd5c5c"  # Red
    INDIGO = "4b0082"  # Purple
    IVORY = "fffff0"  # White
    KHAKI = "f0e68c"  # Yellow
    LAVENDER = "e6e6fa"  # Purple
    LAVENDER_BLUSH = "fff0f5"  # White
    LAWN_GREEN = "7cfc00"  # Green
    LEMON_CHIFFON = "fffacd"  # Yellow
    LIGHT_BLUE = "add8e6"  # Blue
    LIGHT_CORAL = "f08080"  # Red
    LIGHT_CYAN = "e0ffff"  # Blue
    LIGHT_GOLDENROD_YELLOW = "fafad2"  # Yellow
    LIGHT_GRAY = "d3d3d3"  # Gray
    LIGHT_GREEN = "90ee90"  # Green
    LIGHT_PINK = "ffb6c1"  # Pink
    LIGHT_SALMON = "ffa07a"  # Orange
    LIGHT_SEA_GREEN = "20b2aa"  # Green
    LIGHT_SKY_BLUE = "87cefa"  # Blue
    LIGHT_SLATE_GRAY = "778899"  # Gray
    LIGHT_STEEL_BLUE = "b0c4de"  # Blue
    LIGHT_YELLOW = "ffffe0"  # Yellow
    LIME = "00ff00"  # Green
    LIME_GREEN = "32cd32"  # Green
    LINEN = "faf0e6"  # White
    MAGENTA = "ff00ff"  # Purple
    MAROON = "800000"  # Brown
    MEDIUM_AQUA_MARINE = "66cdaa"  # Green
    MEDIUM_BLUE = "0000cd"  # Blue
    MEDIUM_ORCHID = "ba55d3"  # Purple
    MEDIUM_PURPLE = "9370d8"  # Purple
    MEDIUM_SEA_GREEN = "3cb371"  # Green
    MEDIUM_SLATE_BLUE = "7b68ee"  # Blue
    MEDIUM_SPRING_GREEN = "00fa9a"  # Green
    MEDIUM_TURQUOISE = "48d1cc"  # Blue
    MEDIUM_VIOLET_RED = "c71585"  # Pink
    MIDNIGHT_BLUE = "191970"  # Blue
    MINT_CREAM = "f5fffa"  # White
    MISTY_ROSE = "ffe4e1"  # White
    MOCCASIN = "ffe4b5"  # Yellow
    NAVAJO_WHITE = "ffdead"  # Brown
    NAVY = "000080"  # Blue
    OLD_LACE = "fdf5e6"  # White
    OLIVE = "808000"  # Green
    OLIVE_DRAB = "6b8e23"  # Green
    ORANGE = "ffa500"  # Orange
    ORANGE_RED = "ff4500"  # Orange
    ORCHID = "da70d6"  # Purple
    PALE_GOLDEN_ROD = "eee8aa"  # Yellow
    PALE_GREEN = "98fb98"  # Green
    PALE_TURQUOISE = "afeeee"  # Blue
    PALE_VIOLET_RED = "db7093"  # Pink
    PAPAYA_WHIP = "ffefd5"  # Yellow
    PEACH_PUFF = "ffdab9"  # Yellow
    PERU = "cd853f"  # Brown
    PINK = "ffc0cb"  # Pink
    PLUM = "dda0dd"  # Purple
    POWDER_BLUE = "b0e0e6"  # Blue
    PURPLE = "800080"  # Purple
    RED = "ff0000"  # Red
    ROSY_BROWN = "bc8f8f"  # Brown
    ROYAL_BLUE = "4169e1"  # Blue
    SADDLE_BROWN = "8b4513"  # Brown
    SALMON = "fa8072"  # Red
    SANDY_BROWN = "f4a460"  # Brown
    SEA_GREEN = "2e8b57"  # Green
    SEA_SHELL = "fff5ee"  # White
    SIENNA = "a0522d"  # Brown
    SILVER = "c0c0c0"  # Gray
    SKY_BLUE = "87ceeb"  # Blue
    SLATE_BLUE = "6a5acd"  # Purple
    SLATE_GRAY = "708090"  # Gray
    SNOW = "fffafa"  # White
    SPRING_GREEN = "00ff7f"  # Green
    STEEL_BLUE = "4682b4"  # Blue
    TAN = "d2b48c"  # Brown
    TEAL = "008080"  # Green
    THISTLE = "d8bfd8"  # Purple
    TOMATO = "ff6347"  # Orange
    TURQUOISE = "40e0d0"  # Blue
    VIOLET = "ee82ee"  # Purple
    WHEAT = "f5deb3"  # Brown
    WHITE = "ffffff"  # White
    WHITE_SMOKE = "f5f5f5"  # White
    YELLOW = "ffff00"  # Yellow
    YELLOW_GREEN = "9acd32"  # Green

class Direction(Enum):
    LEFT = 1
    RIGHT = 2
    UP = 3
    DOWN = 4
    CLOCKWISE = 5
    COUNTER_CLOCKWISE = 6

def parse_color(value: str, skip_preset = False) -> Color | ColorPreset:
    try:
        if not skip_preset and value.upper() in ColorPreset.__members__:
            return ColorPreset[value.upper()]
        if value.upper() in HtmlColors.__members__:
            return Color(
                int(HtmlColors[value.upper()][:2], 16) % 256,
                int(HtmlColors[value.upper()][2:4], 16) % 256,
                int(HtmlColors[value.upper()][4:], 16) % 256
            )
        elif len(value) == 6:
            return Color(
                int(value[0:2], 16) % 256,
                int(value[2:4], 16) % 256,
                int(value[4:6], 16) % 256
            )
        elif len(value) == 3:
            return Color(
                int(value[0] * 2, 16) % 256,
                int(value[1] * 2, 16) % 256,
                int(value[2] * 2, 16) % 256
            )
        else:
            raise ValueError(f"Can not construct Color from  '{value}'.")
    except Exception as _:
        raise ValueError(f"Can not construct Color from  '{value}'.")

def parse_numeric(value: str, lower_bound: int, upper_bound: int) -> int:
    try:
        if value[-1] == '%':
            percentage = int(value[:-1])
            if not 0 <= percentage <= 100:
                raise ValueError(f"Percentage value '{value}' is out of range.")
            return round(lower_bound + (percentage * (upper_bound - lower_bound) / 100))
        elif '.' in value:
            float_value = float(value)
            if 0 <= float_value <= 1.0:
                return round(lower_bound + (float_value * (upper_bound - lower_bound)))
            elif 1.0 < float_value <= 100.0:
                return round(lower_bound + (float_value / 100 * (upper_bound - lower_bound)))
            else:
                raise ValueError(f"Float value '{value}' is out of range.")
        else:
            raise ValueError(f"Value '{value}' does not end with '%' and is not a float.")
    except Exception as e:
        raise ValueError(f"Cannot construct numeric value from '{value}': {str(e)}")

def parse_brightness(value: str) -> int:
    try:
        return parse_numeric(value, 0, 1 + BRIGHTNESS_MAX)
    except Exception as _:
        raise ValueError(f"Can not construct brightness from  '{value}'.")

def parse_speed(value: str) -> int:
    try:
        return parse_numeric(value, 0, 1 + SPEED_MAX)
    except Exception as _:
        raise ValueError(f"Can not construct speed from  '{value}'.")

def parse_direction(value: str) -> Direction:
    try:
        return Direction[value.upper()]
    except Exception as _:
        raise ValueError(f"Can not construct direction from  '{value}'.")


class CmdStaticColor(KeyboardCommand):
    name = EffectNames.STATIC_COLOR

    @staticmethod
    def parse_arguments(argv: list[str]) -> 'KeyboardCommand':
        """
        :param argv:  Subsection of arguments from command line
        :return: CmdStaticColor instance, if it's possible to parse the input
        
        Valid options are:
        static_color {color: color}
        static_color {color: color} {brightness: quantity}
        If brightness not supplied 100% is assumed
        """
        if len(argv) == 1:
            color = parse_color(argv[0])
            return CmdStaticColor(color)
        elif len(argv) == 2:
            color = parse_color(argv[0])
            brightness = parse_brightness(argv[1])
            return CmdStaticColor(color, brightness)
        else:
            raise ValueError(f'Can not construct CmdStaticCommand from arguments {argv}.')

    def __init__(self, color: Color | ColorPreset, brightness: int = BRIGHTNESS_MAX):
        self.color = color
        self.brightness = brightness

class CmdFlash3x(KeyboardCommand):
    name = EffectNames.FLASH_3X

    @staticmethod
    def parse_arguments(argv: list[str]) -> 'KeyboardCommand':
        """
        :param argv:  Subsection of arguments from command line
        :return: CmdFlash3x instance, if it's possible to parse the input
        
        Valid options are:
        flash_3x {color: color}
        flash_3x {color: color} {brightness: quantity}
        If brightness not supplied 100% is assumed.
        This command doesn't support preset colors because the device doesn't interpret them
        in a standard fashion. Therefore, this command will always send two USB transfers.
        """
        if len(argv) == 1:
            c = parse_color(argv[0], skip_preset = True)
            return CmdFlash3x(c)
        elif len(argv) == 2:
            c = parse_color(argv[0], skip_preset = True)
            brightness = parse_brightness(argv[1])
            return CmdFlash3x(c, brightness)
        else:
            raise ValueError(f'Can not construct CmdFlash3x from arguments {argv}.')

    def __init__(self, color: Color, brightness: int = BRIGHTNESS_MAX):
        self.color = color
        self.brightness = brightness

class CmdBreathing(KeyboardCommand):
    name = EffectNames.FLASH_3X

    @staticmethod
    def parse_arguments(argv: list[str]) -> 'KeyboardCommand':
        """
        :param argv:  Subsection of arguments from command line
        :return: CmdBreathing instance, if it's possible to parse the input.

        Valid options are:  
        breathing {color: color}  
        breathing {color: color} {brightness: quantity}  
        breathing {color: color} {brightness: quantity} {speed: quantity}  
        If brightness or speed is not supplied 100% is assumed.
        This might send one or two USB transfers depending on the choice of color.
        If it is one of the few preset colors, or a random one is requested, only one transfer
        is required.
        """
        if len(argv) == 1:
            c = parse_color(argv[0])
            return CmdBreathing(c)
        elif len(argv) == 2:
            c = parse_color(argv[0])
            brightness = parse_brightness(argv[1])
            return CmdBreathing(c, brightness)
        elif len(argv) == 3:
            c = parse_color(argv[0])
            brightness = parse_brightness(argv[1])
            speed = parse_speed(argv[2])
            return CmdBreathing(c, brightness, speed)
        else:
            raise ValueError(f'Can not construct CmdBreathing from arguments {argv}.')

    def __init__(self, color: Color, brightness: int = BRIGHTNESS_MAX, speed: int = SPEED_MAX):
        self.color = color
        self.brightness = brightness
        self.speed = speed

class CmdWave(KeyboardCommand):
    name = EffectNames.FLASH_3X

    @staticmethod
    def parse_arguments(argv: list[str]) -> 'KeyboardCommand':
        """
        :param argv:  Subsection of arguments from command line
        :return: CmdWave instance, if it's possible to parse the input.

        Valid options are:  
        wave {brightness: quantity}  
        wave {brightness: quantity} {speed: quantity}  
        If brightness or speed is not supplied 100% is assumed.
        This might send one or two USB transfers depending on the choice of color.
        If it is one of the few preset colors, or a random one is requested, only one transfer
        is required.
        """
        if not argv:
            return CmdWave()
        elif len(argv) == 1:
            brightness = parse_brightness(argv[0])
            return CmdWave(brightness)
        elif len(argv) == 2:
            brightness = parse_brightness(argv[0])
            speed = parse_speed(argv[1])
            return CmdWave(brightness, speed)
        else:
            raise ValueError(f'Can not construct CmdWave from arguments {argv}.')

    def __init__(self, brightness: int = BRIGHTNESS_MAX, speed: int = SPEED_MAX):
        self.color = color
        self.brightness = brightness
        self.speed = speed

class CmdNeon(KeyboardCommand):
    name = EffectNames.FLASH_3X

    @staticmethod
    def parse_arguments(argv: list[str]) -> 'KeyboardCommand':
        """
        :param argv:  Subsection of arguments from command line
        :return: CmdNeon instance, if it's possible to parse the input.

        Valid options are:  
        neon {brightness: quantity}  
        neon {brightness: quantity} {speed: quantity}  
        neon {brightness: quantity} {speed: quantity} {direction: quantity}
        If brightness or speed is not supplied, 100% is assumed.
        I speed is not supplied 'right' is assumed.
        """
        if not argv:
            return CmdNeon()
        elif len(argv) == 1:
            brightness = parse_brightness(argv[0])
            return CmdNeon(brightness)
        elif len(argv) == 2:
            brightness = parse_brightness(argv[0])
            speed = parse_speed(argv[1])
            return CmdNeon(brightness, speed)
        elif len(argv) == 3:
            brightness = parse_brightness(argv[0])
            speed = parse_speed(argv[1])
            direction = parse_speed(argv[2])
            return CmdNeon(brightness, speed, direction)
        else:
            raise ValueError(f'Can not construct CmdNeon from arguments {argv}.')

    def __init__(self, brightness: int = BRIGHTNESS_MAX, speed: int = SPEED_MAX, direction: Direction = Direction.RIGHT):
        self.brightness = brightness
        self.speed = speed
        self.direction = direction

class CmdSnake(KeyboardCommand):
    name = EffectNames.FLASH_3X

    @staticmethod
    def parse_arguments(argv: list[str]) -> 'KeyboardCommand':
        """
        :param argv:  Subsection of arguments from command line
        :return: CmdSnake instance, if it's possible to parse the input.

        Valid options are:  
        snake {color: color}  
        snake {color: color} {brightness: quantity}  
        snake {color: color} {brightness: quantity} {speed: quantity}  
        If brightness or speed is not supplied 100% is assumed.
        This might send one or two USB transfers depending on the choice of color.
        If it is one of the few preset colors, or a random one is requested, only one transfer
        is required.
        """
        if len(argv) == 1:
            c = parse_color(argv[0])
            return CmdSnake(c)
        elif len(argv) == 2:
            c = parse_color(argv[0])
            brightness = parse_brightness(argv[1])
            return CmdSnake(c, brightness)
        elif len(argv) == 3:
            c = parse_color(argv[0])
            brightness = parse_brightness(argv[1])
            speed = parse_speed(argv[2])
            return CmdSnake(c, brightness, speed)
        else:
            raise ValueError(f'Can not construct CmdSnake from arguments {argv}.')

    def __init__(self, color: Color, brightness: int = BRIGHTNESS_MAX, speed: int = SPEED_MAX):
        self.color = color
        self.brightness = brightness
        self.speed = speed

class CmdRipple(KeyboardCommand):
    name = EffectNames.FLASH_3X

    @staticmethod
    def parse_arguments(argv: list[str]) -> 'KeyboardCommand':
        """
        :param argv:  Subsection of arguments from command line
        :return: CmdRipple instance, if it's possible to parse the input.

        Valid options are:  
        ripple {color: color}  
        ripple {color: color} {brightness: quantity}  
        ripple {color: color} {brightness: quantity} {speed: quantity}  
        If brightness or speed is not supplied 100% is assumed.
        This might send one or two USB transfers depending on the choice of color.
        If it is one of the few preset colors, or a random one is requested, only one transfer
        is required.
        """
        if len(argv) == 1:
            c = parse_color(argv[0])
            return CmdRipple(c)
        elif len(argv) == 2:
            c = parse_color(argv[0])
            brightness = parse_brightness(argv[1])
            return CmdRipple(c, brightness)
        elif len(argv) == 3:
            c = parse_color(argv[0])
            brightness = parse_brightness(argv[1])
            speed = parse_speed(argv[2])
            return CmdRipple(c, brightness, speed)
        else:
            raise ValueError(f'Can not construct CmdRipple from arguments {argv}.')

    def __init__(self, color: Color, brightness: int = BRIGHTNESS_MAX, speed: int = SPEED_MAX):
        self.color = color
        self.brightness = brightness
        self.speed = speed

class CmdRain(KeyboardCommand):
    name = EffectNames.FLASH_3X

    @staticmethod
    def parse_arguments(argv: list[str]) -> 'KeyboardCommand':
        """
        :param argv:  Subsection of arguments from command line
        :return: CmdRain instance, if it's possible to parse the input.

        Valid options are:  
        rain {color: color}  
        rain {color: color} {brightness: quantity}  
        rain {color: color} {brightness: quantity} {speed: quantity}  
        If brightness or speed is not supplied 100% is assumed.
        This might send one or two USB transfers depending on the choice of color.
        If it is one of the few preset colors, or a random one is requested, only one transfer
        is required.
        """
        if len(argv) == 1:
            c = parse_color(argv[0])
            return CmdRain(c)
        elif len(argv) == 2:
            c = parse_color(argv[0])
            brightness = parse_brightness(argv[1])
            return CmdRain(c, brightness)
        elif len(argv) == 3:
            c = parse_color(argv[0])
            brightness = parse_brightness(argv[1])
            speed = parse_speed(argv[2])
            return CmdRain(c, brightness, speed)
        else:
            raise ValueError(f'Can not construct CmdRain from arguments {argv}.')

    def __init__(self, color: Color, brightness: int = BRIGHTNESS_MAX, speed: int = SPEED_MAX):
        self.color = color
        self.brightness = brightness
        self.speed = speed

class CmdLightning(KeyboardCommand):
    name = EffectNames.FLASH_3X

    @staticmethod
    def parse_arguments(argv: list[str]) -> 'KeyboardCommand':
        """
        :param argv:  Subsection of arguments from command line
        :return: CmdLightning instance, if it's possible to parse the input.

        Valid options are:  
        lightning {color: color}  
        lightning {color: color} {brightness: quantity}  
        lightning {color: color} {brightness: quantity} {speed: quantity}  
        If brightness or speed is not supplied 100% is assumed.
        This might send one or two USB transfers depending on the choice of color.
        If it is one of the few preset colors, or a random one is requested, only one transfer
        is required.
        """
        if len(argv) == 1:
            c = parse_color(argv[0])
            return CmdLightning(c)
        elif len(argv) == 2:
            c = parse_color(argv[0])
            brightness = parse_brightness(argv[1])
            return CmdLightning(c, brightness)
        elif len(argv) == 3:
            c = parse_color(argv[0])
            brightness = parse_brightness(argv[1])
            speed = parse_speed(argv[2])
            return CmdLightning(c, brightness, speed)
        else:
            raise ValueError(f'Can not construct CmdLightning from arguments {argv}.')

    def __init__(self, color: Color, brightness: int = BRIGHTNESS_MAX, speed: int = SPEED_MAX):
        self.color = color
        self.brightness = brightness
        self.speed = speed

class CmdSpot(KeyboardCommand):
    name = EffectNames.FLASH_3X

    @staticmethod
    def parse_arguments(argv: list[str]) -> 'KeyboardCommand':
        """
        :param argv:  Subsection of arguments from command line
        :return: CmdSpot instance, if it's possible to parse the input.

        Valid options are:  
        spot {color: color}  
        spot {color: color} {brightness: quantity}  
        spot {color: color} {brightness: quantity} {speed: quantity}  
        If brightness or speed is not supplied 100% is assumed.
        This might send one or two USB transfers depending on the choice of color.
        If it is one of the few preset colors, or a random one is requested, only one transfer
        is required.
        """
        if len(argv) == 1:
            c = parse_color(argv[0])
            return CmdSpot(c)
        elif len(argv) == 2:
            c = parse_color(argv[0])
            brightness = parse_brightness(argv[1])
            return CmdSpot(c, brightness)
        elif len(argv) == 3:
            c = parse_color(argv[0])
            brightness = parse_brightness(argv[1])
            speed = parse_speed(argv[2])
            return CmdSpot(c, brightness, speed)
        else:
            raise ValueError(f'Can not construct CmdSpot from arguments {argv}.')

    def __init__(self, color: Color, brightness: int = BRIGHTNESS_MAX, speed: int = SPEED_MAX):
        self.color = color
        self.brightness = brightness
        self.speed = speed

class CmdStars(KeyboardCommand):
    name = EffectNames.FLASH_3X

    @staticmethod
    def parse_arguments(argv: list[str]) -> 'KeyboardCommand':
        """
        :param argv:  Subsection of arguments from command line
        :return: CmdStars instance, if it's possible to parse the input.

        Valid options are:  
        stars {color: color}  
        stars {color: color} {brightness: quantity}  
        stars {color: color} {brightness: quantity} {speed: quantity}  
        If brightness or speed is not supplied 100% is assumed.
        This might send one or two USB transfers depending on the choice of color.
        If it is one of the few preset colors, or a random one is requested, only one transfer
        is required.
        """
        if len(argv) == 1:
            c = parse_color(argv[0])
            return CmdStars(c)
        elif len(argv) == 2:
            c = parse_color(argv[0])
            brightness = parse_brightness(argv[1])
            return CmdStars(c, brightness)
        elif len(argv) == 3:
            c = parse_color(argv[0])
            brightness = parse_brightness(argv[1])
            speed = parse_speed(argv[2])
            return CmdStars(c, brightness, speed)
        else:
            raise ValueError(f'Can not construct CmdStars from arguments {argv}.')

    def __init__(self, color: Color, brightness: int = BRIGHTNESS_MAX, speed: int = SPEED_MAX):
        self.color = color
        self.brightness = brightness
        self.speed = speed

class CmdFireball(KeyboardCommand):
    name = EffectNames.FLASH_3X

    @staticmethod
    def parse_arguments(argv: list[str]) -> 'KeyboardCommand':
        """
        :param argv:  Subsection of arguments from command line
        :return: CmdFireball instance, if it's possible to parse the input.

        Valid options are:  
        fireball {color: color}  
        fireball {color: color} {brightness: quantity}  
        fireball {color: color} {brightness: quantity} {speed: quantity}  
        If brightness or speed is not supplied 100% is assumed.
        This might send one or two USB transfers depending on the choice of color.
        If it is one of the few preset colors, or a random one is requested, only one transfer
        is required.
        """
        if len(argv) == 1:
            c = parse_color(argv[0])
            return CmdFireball(c)
        elif len(argv) == 2:
            c = parse_color(argv[0])
            brightness = parse_brightness(argv[1])
            return CmdFireball(c, brightness)
        elif len(argv) == 3:
            c = parse_color(argv[0])
            brightness = parse_brightness(argv[1])
            speed = parse_speed(argv[2])
            return CmdFireball(c, brightness, speed)
        else:
            raise ValueError(f'Can not construct CmdFireball from arguments {argv}.')

    def __init__(self, color: Color, brightness: int = BRIGHTNESS_MAX, speed: int = SPEED_MAX):
        self.color = color
        self.brightness = brightness
        self.speed = speed

class CmdSnow(KeyboardCommand):
    name = EffectNames.FLASH_3X

    @staticmethod
    def parse_arguments(argv: list[str]) -> 'KeyboardCommand':
        """
        :param argv:  Subsection of arguments from command line
        :return: CmdSnow instance, if it's possible to parse the input.

        Valid options are:  
        snow {color: color}  
        snow {color: color} {brightness: quantity}  
        snow {color: color} {brightness: quantity} {speed: quantity}  
        If brightness or speed is not supplied 100% is assumed.
        This might send one or two USB transfers depending on the choice of color.
        If it is one of the few preset colors, or a random one is requested, only one transfer
        is required.
        """
        if len(argv) == 1:
            c = parse_color(argv[0])
            return CmdSnow(c)
        elif len(argv) == 2:
            c = parse_color(argv[0])
            brightness = parse_brightness(argv[1])
            return CmdSnow(c, brightness)
        elif len(argv) == 3:
            c = parse_color(argv[0])
            brightness = parse_brightness(argv[1])
            speed = parse_speed(argv[2])
            return CmdSnow(c, brightness, speed)
        else:
            raise ValueError(f'Can not construct CmdSnow from arguments {argv}.')

    def __init__(self, color: Color, brightness: int = BRIGHTNESS_MAX, speed: int = SPEED_MAX):
        self.color = color
        self.brightness = brightness
        self.speed = speed

class CmdHeartbeat(KeyboardCommand):
    name = EffectNames.FLASH_3X

    @staticmethod
    def parse_arguments(argv: list[str]) -> 'KeyboardCommand':
        """
        :param argv:  Subsection of arguments from command line
        :return: CmdHeartbeat instance, if it's possible to parse the input.

        Valid options are:  
        heartbeat {color: color}  
        heartbeat {color: color} {brightness: quantity}  
        heartbeat {color: color} {brightness: quantity} {speed: quantity}  
        If brightness or speed is not supplied 100% is assumed.
        This might send one or two USB transfers depending on the choice of color.
        If it is one of the few preset colors, or a random one is requested, only one transfer
        is required.
        """
        if len(argv) == 1:
            c = parse_color(argv[0])
            return CmdHeartbeat(c)
        elif len(argv) == 2:
            c = parse_color(argv[0])
            brightness = parse_brightness(argv[1])
            return CmdHeartbeat(c, brightness)
        elif len(argv) == 3:
            c = parse_color(argv[0])
            brightness = parse_brightness(argv[1])
            speed = parse_speed(argv[2])
            return CmdHeartbeat(c, brightness, speed)
        else:
            raise ValueError(f'Can not construct CmdHeartbeat from arguments {argv}.')

    def __init__(self, color: Color, brightness: int = BRIGHTNESS_MAX, speed: int = SPEED_MAX):
        self.color = color
        self.brightness = brightness
        self.speed = speed

ARGUMENTS_TO_COMMANDS = {
    EffectNames.STATIC_COLOR.value: CmdStaticColor,
    EffectNames.FLASH_3X.value: CmdFlash3x,
    EffectNames.BREATHING.value: CmdBreathing,
    EffectNames.WAVE.value: CmdWave,
}
